;
; static char sccs_id[] = "@(#)fchk16.asm	1.9  1/13/95  10:10:58";
;
; This code is being published by Intel to users of the Pentium(tm)
; processor.  Recipients are authorized to copy, modify, compile, use and
; distribute the code.
;
; Intel makes no warranty of any kind with regard to this code, including
; but not limited to, implied warranties or merchantability and fitness for
; a particular purpose. Intel assumes no responsibility for any errors that
; may appear in this code.
;
; No patent licenses are granted, express or implied.
;
;
;  The following code is a PRELIMINARY IMPLEMENTATION of a
;  software patch for the floating point divide instructions.
;
;

.386
.387
                name    fdiv_patch
_DATA   SEGMENT DWORD USE16 PUBLIC 'DATA'

fdiv_risc_table	DB	0, 1, 0, 0, 4, 0, 0, 7, 0, 0, 10, 0, 0, 13, 0, 0
fdiv_scale_1	DD	03f700000h		;0.9375
fdiv_scale_2	DD	03f880000h		;1.0625
one_shl_63	DD	05f000000h

;
;  This is the flag that is tested by all the compiler branches
;  before the divide instruction.
;
;	0 = unknown whether running on a part with the flaw
;	1 = running on a part with no flaw
;	-1= running on a part with the flaw
;

	public	_fdiv_chk_flag
_fdiv_chk_flag 	DW	0h

;
;  The following are initializers for known values that cause
;  the divide problem.
;

test_dividend	DB	0,0,0,0c0h,07eh,01h,050h,041h	 ; x = 4195835.0
test_divisor	DB	0,0,0,080h,0ffh,0ffh,047h,041h   ; y = 3145727.0
one		DB	0,0,0,0,0,0,0f0h,03fh

_DATA   ENDS

;
;  Stack variables for divide routines.
;

DENOM		EQU	0
NUMER		EQU	12
PATCH_CW 	EQU	28
PREV_CW		EQU	24
STACK_SIZE	EQU	32

MAIN_DENOM	EQU	0
MAIN_NUMER	EQU	12



SPILL_SIZE	EQU	12
MEM_OPERAND	EQU	10
SPILL_MEM_OPERAND	EQU	22

;
;  Stack frame locations for MS C/C++
;  Single real is pushed as 4 bytes, Double real is pushed as 8 bytes and
;  Long Double (extended) real is pushed as 10 bytes. Needs to be changed the
;  following definitions for different calling conventions.
;
;

;
; Definition for Single Real
;
SINGLE1		EQU	6  + STACK_SIZE
SINGLE2		EQU	10 + STACK_SIZE

;
; Definition for Double Real
;
DOUBLE1		EQU	6  + STACK_SIZE
DOUBLE2		EQU	14 + STACK_SIZE

;
; Definition for Long Double Real (Extended)
;
LDOUBLE1	EQU	6  + STACK_SIZE
LDOUBLE2	EQU	16 + STACK_SIZE

SRESULT		EQU	14 + STACK_SIZE
DRESULT		EQU	22 + STACK_SIZE

ONESMASK	EQU	0e00h


_TEXT   SEGMENT PARA USE16 PUBLIC 'CODE'
_TEXT   ENDS

_DATA   SEGMENT DWORD USE16 PUBLIC 'DATA'
_DATA   ENDS

CONST   SEGMENT DWORD USE16 PUBLIC 'CONST'
CONST   ENDS

_BSS    SEGMENT DWORD USE16 PUBLIC 'BSS'
_BSS    ENDS

DGROUP  GROUP CONST, _BSS, _DATA



_TEXT   SEGMENT   PARA USE16 PUBLIC 'CODE'


	assume cs:_TEXT, ds:DGROUP, ss:DGROUP


					; In this implementation the
					; fdiv_main_routine is called,
					; therefore all the stack frame
					; locations are adjusted for the
					; return pointer.

fdiv_main_routine PROC	NEAR

	fld     tbyte ptr [bp+MAIN_NUMER]    ; load the numerator
	fld	tbyte ptr [bp+MAIN_DENOM]	; load the denominator
retry:

;  The following three lines test for denormals and zeros.
;  A denormal or zero has a 0 in the explicit digit to the left of the
;  binary point.  Since that bit is the high bit of the word, adding
;  it to itself will produce a carry if and only if the number is not
;  denormal or zero.
;
	mov	ax, [bp+MAIN_DENOM+6]	; get mantissa bits 48-64
	add	ax,ax			; shift one's bit into carry
	jnc	denormal		; if no carry, we have a denormal

;  The following three lines test the three bits after the four bit 
;  pattern (1,4,7,a,d).  If these three bits are not all one, then
;  the denominator cannot expose the flaw.  This condition is tested by
;  inverting the bits and testing that all are equal to zero afterward.

	xor	ax, ONESMASK		; invert the bits that must be one
	test	ax, ONESMASK		; test for needed bits
	jz	scale_if_needed		; if all one, it may need scalling
	fdivp	st(1), st		; OK to use the hardware
	ret
	
;
;  Now we test the four bits for one of the five patterns.
;
scale_if_needed:
	shr	ax, 12			; keep first four bits after point
        push    bx                      ; save bx
        mov     bx,ax                   ; bx = index
	cmp	byte ptr fdiv_risc_table[bx], 0	; check for (1,4,7,a,d)
        pop     bx                      ; restore bx
	jnz	divide_scaled
	fdivp	st(1), st		; OK to use the hardware
	ret
	
divide_scaled:
	mov	ax, [bp + MAIN_DENOM+8]	; test denominator exponent
	and	ax, 07fffh              ; if pseudodenormal ensure that only
	jz	invalid_denom		; invalid exception flag is set
	cmp	ax, 07fffh              ; if NaN or infinity  ensure that only
	je	invalid_denom		; invalid exception flag is set
;
;  The following six lines turn off exceptions and set the
;  precision control to 80 bits.  The former is necessary to
;  force any traps to be taken at the divide instead of the scaling
;  code.  The latter is necessary in order to get full precision for
;  codes with incoming 32 and 64 bit precision settings.  If
;  it can be guaranteed that before reaching this point, the underflow
;  exception is masked and the precision control is at 80 bits, these
;  six lines can be omitted.
;
	fnstcw	[bp+PREV_CW]		; save caller's control word
	mov	ax, [bp+PREV_CW]
	or	ax, 033fh		; mask exceptions, pc=80
	and	ax, 0f3ffh		; set rounding mode to nearest
	mov	[bp+PATCH_CW], ax
	fldcw	[bp+PATCH_CW]		; mask exceptions & pc=80

;  The following lines check the numerator exponent before scaling.
;  This in order to prevent undeflow when scaling the numerator,
;  which will cause a denormal exception flag to be set when the
;  actual divide is preformed. This flag would not have been set
;  normally. If there is a risk of underflow, the scale factor is
;  17/16 instead of 15/16.
;
 	mov	ax, [bp+MAIN_NUMER+8]	; test numerator exponent
 	and	ax, 07fffh
 	cmp	ax, 00001h
 	je	small_numer
 
; perform scaling of both numerator and denominator

	fmul	fdiv_scale_1		; scale denominator by 15/16
	fxch
	fmul	fdiv_scale_1		; scale numerator by 15/16
	fxch

;
;  The next line restores the users control word.  If the incoming
;  control word had the underflow exception masked and precision
;  control set to 80 bits, this line can be omitted.
;

	fldcw	[bp+PREV_CW]		; restore caller's control word
	fdivp	st(1), st		; OK to use the hardware
	ret
	
small_numer:
	fmul	fdiv_scale_2		; scale denominator by 17/16
	fxch
	fmul	fdiv_scale_2		; scale numerator by 17/16
	fxch

;
;  The next line restores the users control word.  If the incoming
;  control word had the underflow exception masked and precision
;  control set to 80 bits, this line can be omitted.
;

	fldcw	[bp+PREV_CW]		; restore caller's control word
	fdivp	st(1), st		; OK to use the hardware
	ret

denormal:
	mov	eax, [bp+MAIN_DENOM]	        ; test for whole mantissa == 0
	or	eax, [bp+MAIN_DENOM+4]	; test for whole mantissa == 0
	jnz	denormal_divide_scaled	; denominator is not zero
invalid_denom:				; zero or invalid denominator
	fdivp	st(1), st		; OK to use the hardware
	ret
	
denormal_divide_scaled:
	mov	ax, word ptr[bp + MAIN_DENOM + 8]	; get exponent
	and	ax, 07fffh		; check for zero exponent
	jnz	invalid_denom
;
;  The following six lines turn off exceptions and set the
;  precision control to 80 bits.  The former is necessary to
;  force any traps to be taken at the divide instead of the scaling
;  code.  The latter is necessary in order to get full precision for
;  codes with incoming 32 and 64 bit precision settings.  If
;  it can be guaranteed that before reaching this point, the underflow
;  exception is masked and the precision control is at 80 bits, these
;  six lines can be omitted.
;

	fnstcw	[bp+PREV_CW]		; save caller's control word
	mov	ax, [bp+PREV_CW]
	or	ax, 033fh		; mask exceptions, pc=80
	and	ax, 0f3ffh		; set rounding mode to nearest
	mov	[bp+PATCH_CW], ax
	fldcw	[bp+PATCH_CW]		; mask exceptions & pc=80

	mov	ax, [bp + MAIN_NUMER +8]	; test numerator exponent
	and	ax, 07fffh		; check for denormal numerator
	je	denormal_numer	
	cmp	ax, 07fffh		; NaN or infinity
	je	invalid_numer
	mov	ax, [bp + MAIN_NUMER + 6]	; get bits 48..63 of mantissa
	add	ax, ax			; shift the first bit into carry
	jnc	invalid_numer		; if there is no carry, we have an
					; invalid numer
	jmp	numer_ok

denormal_numer:
	mov	ax, [bp + MAIN_NUMER + 6]	; get bits 48..63 of mantissa
	add	ax, ax			; shift the first bit into carry
	jc	invalid_numer		; if there is a carry, we have an
					; invalid numer
	
numer_ok:
	fxch
	fstp	st			; pop numerator
	fld	st			; make copy of denominator
	fmul	[one_shl_63]		; make denominator not denormal
	fstp	tbyte ptr [bp+MAIN_DENOM]
	fld	tbyte ptr [bp+MAIN_NUMER]	; load numerator
	fxch				; restore proper order
	fwait

;  The next line restores the users control word.  If the incoming
;  control word had the underflow exception masked and precision
;  control set to 80 bits, this line can be omitted.
;
	fldcw	[bp+PREV_CW]		; restore caller's control word
	jmp	retry			; start the whole thing over

invalid_numer:
;
;  The next line restores the users control word.  If the incoming
;  control word had the underflow exception masked and precision
;  control set to 80 bits, this line can be omitted.
;
	fldcw	[bp + PREV_CW]

	fdivp	st(1), st		; use of hardware is OK.
	ret

fdiv_main_routine	ENDP

;
; int fdiv_detect(void)
;
; This detect code assumes that the code is running on a 386 or better
; processor.
;

	public	_FDIV_DETECT
_FDIV_DETECT PROC FAR
	fld	qword ptr [test_dividend]
	fwait
	fdiv	qword ptr [test_divisor] ; no_error = ((x - (x / y)*y) < 1.0
	fwait
	fmul	qword ptr [test_divisor]
	fwait
	fsubr	qword ptr [test_dividend]
	fwait
	fcomp	qword ptr [one]
	fwait
	fstsw	ax
	fwait
	and	ax, 0100h              ; 0: uninitialized
	shr	ax, 7
	sub	ax, 1
	mov	_fdiv_chk_flag, ax
	ret
	
_FDIV_DETECT ENDP

;
;  The calling convention is FDIV_CHK(float,float,float*)
;
;  This routine assumes a calling convention of single real being pushed as
;  4 bytes.
;

	public  _FDIV_CHK
_FDIV_CHK     PROC    FAR
        push    bp
	sub     sp, STACK_SIZE
        mov     bp,sp
	fld	dword ptr [bp+SINGLE1]
	fstp	tbyte ptr [bp+NUMER]
	fld	dword ptr [bp+SINGLE2]
	fstp	tbyte ptr [bp+DENOM]
	call	fdiv_main_routine 
        mov     ax,es                   ; save es:bx in eax
        shl     eax,16                  ; ..
        mov     ax,bx                   ; ..
        les     bx,[bp+SRESULT]         ; es:bx -> result
	fstp	dword ptr es:[bx]       ; store fdiv result
        mov     bx,ax                   ; restore es:bx
        shr     eax,16                  ; ..
        mov     es,ax                   ; ..
	add	sp, STACK_SIZE
        pop     bp
	ret
_FDIV_CHK     ENDP

;
;  The calling convention is DIV_CHK(double,double,double*)
;
;  This routine assumes a calling convention of double real being pushed as
;  8 bytes.
;
	public  _DIV_CHK
_DIV_CHK     PROC    FAR
        push    bp
	sub     sp, STACK_SIZE
        mov     bp,sp
	fld	qword ptr [bp+DOUBLE1]
	fstp	tbyte ptr [bp+NUMER]
	fld	qword ptr [bp+DOUBLE2]
	fstp	tbyte ptr [bp+DENOM]
	call	fdiv_main_routine 
        mov     ax,es                   ; save es:bx in eax
        shl     eax,16                  ; ..
        mov     ax,bx                   ; ..
        les     bx,[bp+DRESULT]         ; es:bx -> result
	fstp	qword ptr es:[bx]       ; store fdiv result
        mov     bx,ax                   ; restore es:bx
        shr     eax,16                  ; ..
        mov     es,ax                   ; ..
	add	sp, STACK_SIZE
        pop     bp
	ret
_DIV_CHK     ENDP

;
;  The calling convention is LDIV_CHK(long double, long double) and
;  returns a long double, ie, result is in th top of FP Stack (ST).
;
;  This routine assumes a calling convention of long double real being 
;  pushed as  10 bytes.
;
	public  _LDIV_CHK
_LDIV_CHK     PROC    FAR
	push	bp
	sub     sp, STACK_SIZE
	mov	bp,sp
	fld	tbyte ptr [bp+LDOUBLE1]
	fstp	tbyte ptr [bp+NUMER]
	fld	tbyte ptr [bp+LDOUBLE2]
	fstp	tbyte ptr [bp+DENOM]
	call	fdiv_main_routine 
	add	sp, STACK_SIZE
	pop	bp
	ret
_LDIV_CHK     ENDP

_TEXT   ENDS

	end
